@@ -5,15 +5,17 @@ top-level element. That dictionary is merged from global- and node-level paramet
 '''
 
 import typing
+import re
 
 from box import Box
 
 from .. import common
 from .. import data
+from .. import modules
+from ..modules import bgp
+from ..data.validate import must_be_dict,must_be_list,must_be_string,validate_attributes
 from . import nodes
 
-group_attr = [ 'members','vars','config','node_data','device','module' ]
-
 '''
 Return members of the specified group. Recurse through child groups if needed
 '''
@@ -23,13 +25,13 @@ def group_members(topology: Box, group: str, count: int = 0) -> list:
     common.error(
       f'Internal error: unknown group {group}',
       common.IncorrectValue,
-      'groups')                    
-    return []                      
+      'groups')
+    return []
 
   if count > 99:                    # pragma: no cover (impossible to get here, recursive groups are checked elsewhere)
     common.fatal(
       'Recursive group definition, aborting',
-      'groups')                    
+      'groups')
 
   for m in topology.groups[group].members:
     if m in topology.nodes:
@@ -46,8 +48,7 @@ def check_group_data_structure(topology: Box) -> None:
   if not 'groups' in topology:
     topology.groups = Box({},default_box=True,box_dots=True)
 
-  if not isinstance(topology.groups,dict):
-    common.error('Groups topology-level element should be a dictionary of groups')
+  if must_be_dict(topology,'groups','topology',create_empty=True,module='groups') is None:
     return
 
   '''
@@ -65,41 +66,77 @@ def check_group_data_structure(topology: Box) -> None:
   '''
   Sanity checks on global group data
   '''
+
+  list_of_modules = modules.list_of_modules(topology)
+  group_attr = topology.defaults.attributes.group
+  providers = list(topology.defaults.providers.keys())
   for grp,gdata in topology.groups.items():
-    if not isinstance(gdata,Box):
-      common.error('Group definition should be a dictionary: %s' % grp)
+    if must_be_dict(topology.groups,grp,'topology.groups',create_empty=True,module='groups') is None:
       continue
 
+    gpath=f'topology.groups.{grp}'
+    g_modules = gdata.get('module',[])
+    if g_modules:                          # Modules specified in the group -- we know what these nodes will use
+      gm_source = 'group'
+    else:
+      gm_source = 'topology'
+      g_modules = topology.get('module',[])
+
+    validate_attributes(
+      data=gdata,
+      topology=topology,
+      data_path=gpath,
+      data_name='group',
+      attr_list=[ 'group','node' ],
+      module='groups',
+      modules=g_modules,
+      module_source=gm_source,
+      extra_attributes=providers)          # Allow provider-specific settings (not checked at the moment)
+
     if not 'members' in gdata:
       gdata.members = []
 
     if grp == 'all' and gdata.members:
       common.error('Group "all" should not have explicit members')
 
-    if 'vars' in gdata:
-      if not isinstance(gdata.vars,dict):
-        common.error('Group variables must be a dictionary: %s' % grp)
+    must_be_dict(gdata,'vars',gpath,create_empty=False,module='groups')
+    must_be_dict(gdata,'node_data',gpath,create_empty=False,module='groups')
+    must_be_list(gdata,'config',gpath,create_empty=False,module='groups')
+    must_be_list(gdata,'module',gpath,create_empty=False,module='groups',valid_values=list_of_modules)
+    must_be_string(gdata,'device',gpath,module='groups',valid_values=list(topology.defaults.devices))
 
-    for attr in ('config','module'):
-      if attr in gdata:
-        data.must_be_list(gdata,attr,f'groups.{grp}')
+    if 'node_data' in gdata:                 # Validate node_data attributes (if any)
+      validate_attributes(
+        data=gdata.node_data,
+        topology=topology,
+        data_path=f'{gpath}.node_data',
+        data_name='node',
+        attr_list=[ 'node' ],
+        module='groups',
+        modules=g_modules,
+        module_source=gm_source,
+        extra_attributes=providers)          # Allow provider-specific settings (not checked at the moment)
 
-    for attr in ('device'):
-      data.must_be_string(gdata,attr,f'groups.{grp}')
+      for k in ('module','device'):          # Check that the 'module' or 'device' attributes are not in node_data
+        if k in gdata.node_data:
+          common.error(
+            f'Cannot use attribute {k} in node_data in group {grp}, set it as a group attribute',
+            common.IncorrectValue,
+            'groups')
 
-    for k in gdata.keys():
-      if not k in group_attr:
-        common.error('Unknown attribute "%s" in group %s' % (k,grp))
+    for k in list(gdata.keys()):             # Then move (validated) group node attributes into node_data
+      if k in group_attr:
+        continue
+      gdata.node_data[k] = gdata[k]
+      gdata.pop(k,None)
 
-    if not isinstance(gdata.members,list):
-      common.error('Group members must be a list of nodes or groups: %s' % grp)
+    if must_be_list(gdata,'members',gpath,create_empty=False,module='groups') is None:
       continue
 
     for n in gdata.members:
       if not n in topology.nodes and not n in topology.groups:
         common.error('Member %s of group %s is not a valid node or group name' % (n,grp))
 
-
 '''
 Add node-level group settings to global groups
 '''
@@ -108,8 +145,7 @@ def add_node_level_groups(topology: Box) -> None:
     if not 'group' in n:
       continue
 
-    if isinstance(n.group,str):                           # Node group should be a list
-      n.group = [ n.group ]                               # Convert a string value into a single-element list
+    must_be_list(n,'group',f'nodes.{name}')
 
     for grpname in n.group:
       if not grpname in topology.groups:
@@ -125,8 +161,8 @@ Check recursive group definitions
 def check_recursive_chain(topology: Box, chain: list, group: str) -> typing.Optional[list]:
   if not group in topology.groups: # pragma: no cover (if we ever get here we're seriously messed up)
     common.fatal(
-      'Internal error: unknown group in check_recursive_chain') 
-    return None  
+      'Internal error: unknown group in check_recursive_chain')
+    return None
 
   chain = chain + [ group ]
   for m in topology.groups[group].members:
@@ -167,58 +203,47 @@ def copy_group_device_module(topology: Box) -> None:
   for grp in reverse_topsort(topology):
     gdata = topology.groups[grp]
     if 'device' in gdata or 'module' in gdata:
+      if common.debug_active('groups'):
+        print(f'Setting device/module for group {grp}')
       g_members = group_members(topology,grp)
-      if g_members:
-        for name,ndata in topology.nodes.items():
-          if name in g_members:
-            for attr in ('device','module'):
-              if attr in gdata and not attr in ndata:
-                ndata[attr] = gdata[attr]
-      else:
+      if not g_members:
         common.error(
           f'Cannot use "module" or "device" attribute on in group {grp} that has no direct or indirect members',
           common.IncorrectValue,
           'groups')
+        continue
 
-'''
-Utility function: merge group value into node value. Deals with scalars and dicts
-'''
-def merge_node_data(ndata: Box, k: str, v: typing.Any, node_name: str, group_name: str) -> None:
-  if not k in ndata:
-    ndata[k] = v
-  if isinstance(ndata[k],dict):
-    if isinstance(v,dict):
-      ndata[k] = ndata[k] + v
-    else:
-      common.error(
-        f'Cannot merge non-dictionary node_data {k} from group {group_name} into node {node_name}',
-        common.IncorrectValue,
-        'groups')
+      for name,ndata in topology.nodes.items():
+        if not name in g_members:
+          continue
+
+        for attr in ('device','module'):
+          if attr in gdata and not attr in ndata:
+            ndata[attr] = gdata[attr]
+            if common.debug_active('groups'):
+              print(f'... setting {attr} on {name} to {gdata[attr]}')
 
 '''
 Copy node data from group into group members
 '''
-def copy_group_node_data(topology: Box) -> None:
+def copy_group_node_data(topology: Box,pfx: str) -> None:
   for grp in reverse_topsort(topology):
+    if not grp.startswith(pfx):                                       # Skip groups that don't match the current prefix (ex: BGP autogroups)
+      continue
     gdata = topology.groups[grp]
-    if 'node_data' in gdata:
-      node_data_keys  = set(gdata.node_data.keys())                   # You cannot set device type or node modules with
-      node_data_wrong = False                                         # ... node data, those attributes have to be set
-      for k in ('module','device'):                                   # ... at group level
-        if k in node_data_keys:
-          common.error(
-            f'Cannot use attribute {k} in node_data in group {grp}, set it as a group attribute',
-            common.IncorrectValue,
-            'groups')
-          node_data_wrong = True
-      if node_data_wrong:
+    if not 'node_data' in gdata:                                      # No group data, skip
+      continue
+
+    g_members = group_members(topology,grp)                           # Get recursive list of members
+    if common.debug_active('groups'):
+      print(f'copy node data {grp}: {gdata.node_data}')
+    for name in g_members:                                            # Iterate over group members
+      if not name in topology.nodes:                                  # Member is not a node, skip it
         continue
 
-      g_members = group_members(topology,grp)
-      for name,ndata in topology.nodes.items():
-        if name in g_members:
-          for k,v in gdata.node_data.items():   # Have to go one level deeper, changing ndata value wouldn't work
-            merge_node_data(ndata,k,v,ndata.name,grp)
+      if common.debug_active('groups'):
+        print(f'... merging node data with {name}')
+      topology.nodes[name] = gdata.node_data + topology.nodes[name]
 
 '''
 Export node_data from groups to topology
@@ -244,20 +269,61 @@ def export_group_node_data(
   for gname,gdata in topology.groups.items():
     #
     # Find groups with node_data dictionaries
-    if data.must_be_dict(gdata,f'node_data.{key}',f'groups.{gname}',module=module,create_empty=False):
-      for obj_name in gdata.node_data[key].keys():
+    if must_be_dict(gdata,f'node_data.{key}',f'groups.{gname}',module=module,create_empty=False):
+      for obj_name in list(gdata.node_data[key].keys()):
+        if gdata.node_data[key][obj_name] is None:
+          gdata.node_data[key][obj_name] = {}
+        obj_data =  gdata.node_data[key][obj_name]
         if not obj_name in topology[key] or topology[key][obj_name] is None:
           topology[key][obj_name] = {}
         for attr in unique_keys:
-          if attr in topology[key][obj_name] and attr in gdata.node_data[key][obj_name] and \
-             topology[key][obj_name][attr] != gdata.node_data[key][obj_name][attr]:     # Unique key present on both ends and not equal
+          if attr in topology[key][obj_name] and attr in obj_data and \
+             topology[key][obj_name][attr] != obj_data[attr]:                          # Unique key present on both ends and not equal
             common.error(
               f'Cannot redefine {key} attribute {attr} for {key}.{obj_name} from node_data in group {gname}',
               common.IncorrectValue,
               module)
         for attr in copy_keys:
-          if attr in gdata.node_data[key][obj_name] and attr not in topology[key][obj_name]:
-            topology[key][obj_name][attr] = gdata.node_data[key][obj_name][attr]
+          if attr in obj_data and attr not in topology[key][obj_name]:
+            topology[key][obj_name][attr] = obj_data[attr]
+
+#
+# create_bgp_autogroups -- create BGP AS groups
+#
+
+def create_bgp_autogroups(topology: Box) -> None:
+  g_module = topology.get('module',[])                          # Global list of modules, could be invalid
+  if not isinstance(g_module,list):                             # Won't deal with incorrectly formatted 'module' attribute here
+    g_module = []                                               # ... just assume it's an empty list
+  g_bgpas = data.get_global_parameter(topology,'bgp.as')        # Try to get the global BGP AS
+  if not data.is_true_int(g_bgpas):                             # ... don't worry if it's not int, someone else will complain
+    g_bgpas = 0
+
+  for gname,gdata in topology.groups.items():                   # Sanity check: BGP autogroups should not have static members
+    if re.match('as\\d+$',gname):
+      if gdata.get('members',None):                             # Well, it's OK to have an empty list of members ;)
+        common.error(
+          f'BGP AS group {gname} should not have static members',
+          common.IncorrectValue,
+          'groups')
+
+  for n_name,n_data in topology.nodes.items():                  # Now iterate over nodes
+    n_module = n_data.get('module',g_module)                    # Get node or global module (global modules haven't been propagated yet)
+    if not isinstance(n_module,list):                           # Node list of modules is insane, someone else will complain
+      continue
+
+    if not 'bgp' in n_module:                                   # Looks like this node does not care about BGP
+      continue
+
+    n_bgpas = data.get_from_box(n_data,'bgp.as') or g_bgpas     # Get node-level or global BGP AS
+    if not n_bgpas:
+      continue
+
+    grpname = f"as{n_bgpas}"                                    # BGP auto-group name
+    if not 'members' in topology.groups[grpname]:               # Set members of a new group to an empty list
+      topology.groups[grpname].members = []                     # ... because we're using Box this also creates an empty group dict
+
+    topology.groups[grpname].members.append(n_name)             # ... and append the node to AS group
 
 #
 # init_groups:
@@ -278,22 +344,11 @@ def init_groups(topology: Box) -> None:
   common.exit_on_error()
 
   copy_group_device_module(topology)
-
-#
-# adjust_groups:
-#
-# * Copy group data into nodes
-#
-# We have to split the group initialization code into two phases
-# due to BGP auto groups.
-#
-def adjust_groups(topology: Box) -> None:
-  copy_group_node_data(topology)
+  copy_group_node_data(topology,'')                 # Copy all group data into nodes (potentially setting bgp.as)
+  bgp.process_as_list(topology)
+  create_bgp_autogroups(topology)                   # Create AS-based groups
+  copy_group_node_data(topology,'as')               # And add group data from 'asxxxx' into nodes
   common.exit_on_error()
-
-  '''
-  Finally, remove 'groups' topology element if it's not needed
-  '''
   if not topology.groups:
     del topology['groups']
 
@@ -318,11 +373,11 @@ def node_config_templates(topology: Box) -> None:
     if not 'config' in topology.groups[group_name]:
       continue
 
-    data.must_be_list(topology.groups[group_name],'config',f'groups.{group_name}')
+    must_be_list(topology.groups[group_name],'config',f'groups.{group_name}')
     g_members = group_members(topology,group_name)
     for name,ndata in topology.nodes.items():
       if name in g_members or group_name == 'all':
-        if not data.must_be_list(ndata,'config',f'nodes.{name}') is None:
+        if not must_be_list(ndata,'config',f'nodes.{name}') is None:
           ndata.config = topology.groups[group_name].config + ndata.config
 
   '''
