@@ -11,6 +11,7 @@ from .. import addressing
 from .. import augment
 from ..providers import _Provider
 from .. import modules
+from .. import quirks
 from ..data import global_vars
 
 def transform_setup(topology: Box) -> None:
@@ -36,14 +37,17 @@ def transform_setup(topology: Box) -> None:
   augment.nodes.augment_node_system_data(topology)
   common.exit_on_error()
 
+  modules.pre_default(topology)
+  common.exit_on_error()
+
+  augment.topology.check_global_elements(topology)
+  augment.nodes.validate(topology)
+
 def transform_data(topology: Box) -> None:
-  addressing.setup(topology,topology.defaults)
+  addressing.setup(topology)
   augment.plugin.execute('pre_transform',topology)
   modules.pre_transform(topology)
 
-  augment.groups.adjust_groups(topology)
-  augment.topology.check_global_elements(topology)
-
   augment.plugin.execute('pre_node_transform',topology)
   modules.pre_node_transform(topology)
   augment.nodes.transform(topology,topology.defaults,topology.pools)
@@ -52,8 +56,11 @@ def transform_data(topology: Box) -> None:
   modules.post_node_transform(topology)
 
   if 'links' in topology:
+    augment.links.validate(topology)
+    common.exit_on_error()
     augment.plugin.execute('pre_link_transform',topology)
     modules.pre_link_transform(topology)
+    common.exit_on_error()
     augment.links.transform(topology.links,topology.defaults,topology.nodes,topology.pools)
     common.exit_on_error()
     augment.plugin.execute('post_link_transform',topology)
@@ -66,6 +73,9 @@ def transform_data(topology: Box) -> None:
   topology.Provider.call("post_transform",topology)
   common.exit_on_error()
 
+  quirks.process_quirks(topology)
+  common.exit_on_error()
+  
   topology.pop('Plugin',None)
   del topology.pools
   del topology.Provider
