@@ -1,15 +1,15 @@
 #
-# VRF module
+# VLAN configuration module
 #
 import typing
 from box import Box
 
-from . import _Module,_routing,get_effective_module_attribute
+from . import _Module,_routing,get_effective_module_attribute,_dataplane
 from .. import common
 from .. import data
-from ..data import global_vars
+from ..data import global_vars,get_from_box,get_global_parameter
+from ..data.validate import validate_attributes
 from .. import addressing
-from ..data import get_from_box,get_global_parameter
 from ..augment import devices,groups
 from ..augment import links
 
@@ -24,63 +24,13 @@ vlan_link_attr: typing.Final[dict] = {
   'trunk' : { 'type' : dict,'vlan': True }
 }
 
-phy_ifattr: typing.Final[list] = ['bridge','ifindex','parentindex','ifname','linkindex','type','vlan','mtu','_selfloop_ifindex'] # Physical interface attributes
-keep_subif_attr: typing.Final[list] = ['vlan','ifindex','ifname','type']    # Keep these attributes on VLAN subinterfaces
-vlan_link_attr_copy: typing.Final[list] = ['role','unnumbered','pool']      # VLAN attributes to copy to member links
-
-#
-# build_vlan_id_set: given an object (topology or node), create a set of VLAN attributes (ID or VNI)
-# that appear in that object.
-#
-def build_vlan_id_set(obj: Box, attr: str, objname: str) -> set:
-  if 'vlans' in obj:
-    if not isinstance(obj.vlans,dict):    # pragma: no cover
-      common.fatal(f'Found a "vlans" setting that is not a dictionary in {objname}','vlan')
-      return set()
-
-    return { v[attr] for v in obj.vlans.values() if isinstance(v,dict) and attr in v and v[attr] is not None and not isinstance(v[attr],bool) }
-  return set()
-
 def populate_vlan_id_set(topology: Box) -> None:
-  vlan_ids = global_vars.get('vlan_ids')
-  vlan_next = global_vars.get('vlan_next')
-
-  for k in ('id','vni'):
-    vlan_ids[k] = build_vlan_id_set(topology,k,'topology')
-
-  vlan_next['id'] = topology.defaults.vlan.start_vlan_id
-  vlan_next['vni'] = topology.defaults.vlan.start_vni
-
-  attr_list = [ 'id' ]
-  if topology.defaults.vlan.auto_vni:
-    attr_list.append('vni')
+  _dataplane.create_id_set('vlan_id')
+  _dataplane.extend_id_set('vlan_id',_dataplane.build_id_set(topology,'vlans','id','topology'))
+  _dataplane.set_id_counter('vlan_id',topology.defaults.vlan.start_vlan_id,4094)
 
   for n in topology.nodes.values():
-    if 'vlans' in n:
-      for k in attr_list:
-        vlan_ids[k] = vlan_ids[k].union(build_vlan_id_set(n,k,n.name))
-
-def get_next_vlan_id(k : str) -> int:
-  vlan_ids = global_vars.get('vlan_ids')
-  vlan_next = global_vars.get('vlan_next')
-
-  if k not in vlan_ids or k not in vlan_next:   # pragma: no cover
-    common.fatal(f'Invalid attribute {k} in get_next_vlan_id call')
-
-  while vlan_next[k] in vlan_ids[k]:
-    vlan_next[k] = vlan_next[k] + 1
-
-  vlan_ids[k].add(vlan_next[k])
-  return vlan_next[k]
-
-#
-# Define a utility function to check whether a VLAN ID or VNI is used to hide internal data structures
-# from modules needing this functionality
-#
-def is_vlan_id_used(vlan_id: int, namespace: str) -> bool:
-  vlan_ids = global_vars.get('vlan_ids')
-
-  return vlan_id in vlan_ids[namespace]
+    _dataplane.extend_id_set('vlan_id',_dataplane.build_id_set(n,'vlans','id',f'nodes.{n.name}'))
 
 #
 # routed_access_vlan: Given a link with access/native VLAN, check if all nodes on the link use routed VLAN
@@ -128,8 +78,6 @@ def interface_vlan_mode(intf: Box, node: Box, topology: Box) -> str:
 # * VLAN forwarding mode
 #
 def validate_vlan_attributes(obj: Box, topology: Box) -> None:
-  vlan_ids = global_vars.get('vlan_ids')
-
   obj_name = 'global VLANs' if obj is topology else obj.name
   obj_path = 'vlans' if obj is topology else f'nodes.{obj.name}.vlans'
   default_fwd_mode = get_from_box(obj,'vlan.mode')                # Get node-wide VLAN forwarding mode
@@ -151,6 +99,15 @@ def validate_vlan_attributes(obj: Box, topology: Box) -> None:
       obj.vlans[vname] = Box({},default_box=True,box_dots=True)
 
     vdata = obj.vlans[vname]
+    validate_attributes(
+      data=vdata,                                     # Validate node data
+      topology=topology,
+      data_path=f'{obj_path}.{vname}',                # Topology path to VLAN attributes
+      data_name=f'VLAN',
+      attr_list=['vlan','link'],                      # We're checking VLAN and link attributes
+      modules=obj.get('module',[]),                   # ... against object modules (node or topology)
+      module_source='topology' if obj is topology else f'nodes.{obj.name}',
+      module='vlans')                                 # Function is called from 'vlans' module
 
     if 'mode' in vdata:                                             # Do we have 'mode' set in the VLAN definition?
       if not vdata.mode in vlan_mode_kwd:                           # ... check the keyword value
@@ -163,7 +120,7 @@ def validate_vlan_attributes(obj: Box, topology: Box) -> None:
         vdata.mode = default_fwd_mode
 
     if not 'id' in vdata:                                           # When VLAN ID is not defined
-      vdata.id = get_next_vlan_id('id')                             # ... take the next free VLAN ID from the list
+      vdata.id = _dataplane.get_next_id('vlan_id')                  # ... take the next free VLAN ID from the list
     if not isinstance(vdata.id,int):                                # Now validate the heck out of VLAN ID
       common.error(f'VLAN ID {vdata.id} for VLAN {vname} in {obj_name} must be an integer',common.IncorrectValue,'vlan')
       continue
@@ -171,33 +128,12 @@ def validate_vlan_attributes(obj: Box, topology: Box) -> None:
       common.error(f'VLAN ID {vdata.id} for VLAN {vname} in {obj_name} must be between 2 and 4094',common.IncorrectValue,'vlan')
       continue
 
-    # Assign VNI only when topology.defaults.vlan.auto_vni is set and there's no VNI in VLAN data
-    #
-    assign_vni = topology.defaults.vlan.auto_vni and not 'vni' in vdata
-    if 'vni' in vdata:
-      if vdata.vni is True:                                         # VNI auto-assignment requested
-        assign_vni = True
-      elif vdata.vni is False:                                      # Explicit request not to assign a VNI
-        vdata.pop('vni')
-      elif not isinstance(vdata.vni,int):                           # VNI is not bool, validate that it's int and in correct range
-        common.error(f'VNI {vdata.vni} for VLAN {vname} in {obj_name} must be an integer',common.IncorrectValue,'vlan')
-        continue
-      elif vdata.vni < 2 or vdata.vni > 16777215:
-        common.error(f'VNI {vdata.vni} for VLAN {vname} in {obj_name} must be between 2 and 16777215',common.IncorrectValue,'vlan')
-        continue
-
-    if assign_vni:                                                  # So far we either have a valid VNI or know whether to auto-assign it
-      vni_default = topology.defaults.vlan.start_vni + vdata.id     # ... try to build VNI from VLAN ID
-      if not is_vlan_id_used(vni_default,'vni'):                    # Is the VNI free?
-        vdata.vni = vni_default                                     # ... great, take it
-        vlan_ids.vni.add(vni_default)                               # ... and add it to the list of used VNIs
-      else:                                                         # Too bad, we had such a great idea but it failed
-        vdata.vni = get_next_vlan_id('vni')                         # ... so take the next available VNI
-
     vlan_pool = [ vdata.pool ] if 'pool' in vdata else []
     vlan_pool.extend(['vlan','lan'])
-    pfx_list = links.augment_link_prefix(vdata,vlan_pool,topology.pools,f'{obj_path}.{vname}')
+    pfx_list = links.assign_link_prefix(vdata,vlan_pool,topology.pools,f'{obj_path}.{vname}')
     vdata.prefix = addressing.rebuild_prefix(pfx_list)
+    if not 'allocation' in vdata.prefix:
+      vdata.prefix.allocation = 'id_based'
 
 """
 check_link_vlan_attributes: check correctness of VLAN link attributes
@@ -365,6 +301,61 @@ def validate_link_vlan_attributes(link: Box,v_attr: Box,topology: Box) -> bool:
                 intf[af] = False                                  # ... make sure it's not connected to the native VLAN subnet
   return link_ok
 
+
+"""
+validate_trunk_vlan_list -- validate that the nodes connected to a VLAN trunk have at least some common VLANs
+
+At this point, the interfaces should contain a fully-blown list of VLAN attributes
+"""
+def validate_trunk_vlan_list(link: Box) -> bool:
+  needs_native = [ intf.node for intf in link.interfaces if not 'vlan' in intf ]
+  has_native   = [ intf.node for intf in link.interfaces if 'vlan' in intf and 'native' in intf.vlan ]
+
+  if common.debug_active('vlan'):
+    print(f'Validate trunk VLAN list -- needs_native: {needs_native} has_native: {has_native}')
+
+  # First check: if we have a non-VLAN node we must have a native VLAN
+  if needs_native and not has_native:
+    common.error(
+      f'Non-VLAN nodes are attached to VLAN trunk links[{link.linkindex}] that has no native VLAN',
+      common.IncorrectValue,
+      'vlan')
+    return False
+
+  # Second check: Every VLAN used by a node must be used by at least one other node
+  has_error = False
+  for o_intf in link.interfaces:                                      # Iterate over link interfaces
+    if not 'vlan' in o_intf:                                          # Skip non-VLAN interfaces (obviously we have a native VLAN by now)
+      continue
+    if not 'trunk' in o_intf.vlan:                                    # a VLAN interface without a trunk attribute is a huge red flag
+      common.fatal('validate_trunk_vlan_list: Found a VLAN node without trunk attribute\n... {link}')
+      return False
+
+    for vname in o_intf.vlan.trunk.keys():                            # OK, now we can iterate over all VLANs in this interfaces' trunk
+      vlan_found = False
+      for i_intf in link.interfaces:                                  # ... and over all other other interfaces
+        if i_intf is o_intf or not 'vlan' in i_intf:                  # ... I did say all OTHER VLAN interfaces, right?
+          continue
+        if not 'trunk' in i_intf.vlan:                                # We'll deal with this abomination in the outer loop
+          continue
+        if vname in i_intf.vlan.trunk:                                # Ah, found a matching VLAN, life is good, stop wasting CPU cycles
+          vlan_found = True
+          break
+
+      if vlan_found:                                                  # By the end of innermost loop we should have an answer to this question
+        continue
+
+      if vname == o_intf.vlan.get('native',None) and needs_native:    # But maybe we found a native VLAN that someone needs?
+        continue
+
+      common.error(
+        f'VLAN {vname} used by node {o_intf.node} on links[{link.linkindex}] is not used by any other node on that link',
+        common.IncorrectValue,
+        'vlan')
+      has_error = True
+
+  return not has_error
+
 """
 copy_vlan_attributes: copy prefix and link type from vlan to link
 """
@@ -436,6 +427,58 @@ def set_link_vlan_prefix(link: Box, v_attr: Box, topology: Box) -> None:
 
   copy_vlan_attributes(link_vlan,topology.nodes[list(node_set)[0]].vlans[link_vlan],link)
 
+"""
+create_vlan_link_data: Create initial link data for a VLAN member link
+
+Used by create_vlan_links and create_loopback_vlan_links
+"""
+def create_vlan_link_data(init: typing.Union[Box,dict],vname: str, parent: typing.Any, topology: Box) -> Box:
+  link_data = Box(init,default_box=True,box_dots=True)
+  link_data.linkindex = topology.links[-1].linkindex + 1
+  link_data.parentindex = parent
+  link_data.vlan.access = vname
+  link_data.vlan_name = vname
+  link_data.type = 'vlan_member'
+  link_data.interfaces = []
+  fix_vlan_mode_attribute(link_data)
+
+  if vname in topology.get('vlans',{}):
+    vdata = topology.vlans[vname]
+    for k in vdata.keys():                              # Copy list-related VLAN attributes
+      if k in topology.defaults.vlan.attributes.vlan_no_propagate:
+        continue
+
+      link_data[k] = vdata[k]
+
+  return link_data
+
+"""
+create_vlan_member_interface: Create interface data for a VLAN mamber link
+
+Used by create_vlan_links and create_loopback_vlan_links
+"""
+def create_vlan_member_interface(
+      init: typing.Union[Box,dict],               # Initial VLAN parameters taken from vlan.trunk
+      vname: str,                                 # VLAN name
+      parent_intf: Box,                           # Parent interface data (used to get node name etc)
+      topology: Box,
+      loop_index: int) -> Box:                    # selfloop_ifindex to support crazy topologies
+
+  intf_data = Box(init,default_box=True,box_dots=True)
+  intf_data.node = parent_intf.node
+  intf_data._selfloop_ifindex = loop_index                  # Used in find_parent_interface to disambiguate self-links
+  intf_data.vlan.access = vname
+
+  if 'mode' in parent_intf.vlan and not get_from_box(intf_data,'vlan.mode'):
+    intf_data.vlan.mode = parent_intf.vlan.mode             # vlan.mode is inherited from trunk dictionary or parent interface
+
+  intf_node = topology.nodes[parent_intf.node]
+  if interface_vlan_mode(intf_data,intf_node,topology) == 'bridge':     # Is this VLAN interface in bridge mode?
+    intf_data.ipv4 = False                                              # ... if so, disable addressing on this interface
+    intf_data.ipv6 = False
+
+  return intf_data
+
 """
 create_vlan_links: Create virtual links for every VLAN in the VLAN trunk
 """
@@ -447,42 +490,23 @@ def create_vlan_links(link: Box, v_attr: Box, topology: Box) -> None:
 
   for vname in sorted(v_attr.trunk.set):
     if vname != native_vlan:           # Skip native VLAN
-      link_data = Box(link.vlan.trunk[vname] or {},default_box=True,box_dots=True)
-      link_data.linkindex = topology.links[-1].linkindex + 1
-      link_data.parentindex = link.linkindex
-      link_data.vlan.access = vname
-      link_data.vlan_name = vname
-      link_data.type = 'vlan_member'
-      link_data.interfaces = []
-      fix_vlan_mode_attribute(link_data)
+      link_data = create_vlan_link_data(link.vlan.trunk[vname] or {},vname,link.linkindex,topology)
 
-      if vname in topology.get('vlans',{}):                 # We need an IP prefix for the VLAN link
-        vdata = topology.vlans[vname]
-        prefix = vdata.prefix                               # Hopefully we can get it from the global VLAN pool
-        for k in vlan_link_attr_copy:                       # ... also copy other link-related VLAN attributes (role, pool)
-          if k in vdata:
-            link_data[k] = vdata[k]
+      prefix = None
+      if vname in topology.get('vlans',{}):
+        prefix = topology.vlans[vname].get('prefix',None)
 
       selfloop_ifindex = 0
       for intf in link.interfaces:
         if 'vlan' in intf and vname in intf.vlan.get('trunk',{}):
-          intf_data = Box(intf.vlan.trunk[vname] or {},default_box=True,box_dots=True)
-          intf_data.node = intf.node
-          intf_data._selfloop_ifindex = selfloop_ifindex     # Used in find_parent_interface to disambiguate self-links
+          intf_data = create_vlan_member_interface(intf.vlan.trunk[vname] or {},vname,intf,topology,selfloop_ifindex)
           selfloop_ifindex = selfloop_ifindex + 1
-          intf_data.vlan.access = vname
           intf_node = topology.nodes[intf.node]
 
-          if 'mode' in intf.vlan and not get_from_box(intf_data,'vlan.mode'):
-            intf_data.vlan.mode = intf.vlan.mode            # vlan.mode is inherited from trunk dictionary or parent interface
-
-          if interface_vlan_mode(intf_data,intf_node,topology) == 'bridge':     # Is this VLAN interface in bridge mode?
-            intf_data.ipv4 = False                                              # ... if so, disable addressing on this interface
-            intf_data.ipv6 = False
-          else:
-            if not prefix:                                  # Still no usable IP prefix? Try to get it from the node VLAN pool
-              if vname in intf_node.get('vlans',{}):
-                prefix = topology.node[intf.node].vlans[vname].prefix
+          # Still no usable IP prefix? Try to get it from the node VLAN pool, but only if VLAN is not in bridge mode
+          if not prefix and vname in intf_node.get('vlans',{}):
+            if interface_vlan_mode(intf_data,intf_node,topology) != 'bridge':
+              prefix = topology.node[intf.node].vlans[vname].get('prefix',None)
 
           link_data.interfaces.append(intf_data)            # Append the interface to vlan link
 
@@ -498,6 +522,49 @@ def create_vlan_links(link: Box, v_attr: Box, topology: Box) -> None:
 
       topology.links.append(link_data)
 
+"""
+create_loopback_vlan_links
+
+Sometimes we have to create a SVI/VLAN interface even though a node has no physical interface participating
+in that VLAN. Typical use cases are VXLAN transport used in VRF-lite scenarios, asymmetric IRB, or VXLAN-to-VXLAN routing
+
+This routine collects node VLAN attachment data from 'vlan_name' attribute set on links by link_pre_transform code,
+identifies missing VLANs, and creates fake links with a single node attached to VLAN access interface.
+"""
+def create_loopback_vlan_links(topology: Box) -> None:
+  node_vlan_list: dict = {}
+  for link_data in topology.links:                                      # Phase 1: collect per-node VLAN attachment data
+    if not 'vlan_name' in link_data:                                    # ... skip non-VLAN links
+      continue
+    for intf in link_data.interfaces:                                   # Now iterate over all interfaces attached to the link
+      if not intf.node in node_vlan_list:                               # Create empty VLAN lists for new nodes
+        node_vlan_list[intf.node] = []
+      if not link_data.vlan_name in node_vlan_list[intf.node]:          # Do we know about this particular VLAN/node attachment?
+        node_vlan_list[intf.node].append(link_data.vlan_name)           # ... not yet, append it to the list
+
+  for n in topology.nodes.values():                                     # Now iterate over nodes
+    if not 'vlans' in n:                                                # Move on if the node has no 'vlans' attribute
+      continue
+    for vname in n.vlans.keys():                                        # Now iterate over VLANs defined on a node
+      if n.name in node_vlan_list and vname in node_vlan_list[n.name]:  # VLAN already present on a node interface, move on
+        continue
+
+      # and now the real work starts: create a fake VLAN link with a single node attached to it
+      if common.debug_active('vlan'):
+        print(f'Creating loopback link for VLAN {vname} on node {n.name}')
+      link_data = create_vlan_link_data({},vname,'loopback',topology)   # Create a vlan_member link with fake parent (nobody should ever use it)
+      prefix = topology.vlans[vname].get('prefix',None)                 # Copy VLAN prefix into link_data
+      if prefix:
+        link_data.prefix = prefix
+
+      # Create interface data using fake parent interface
+      fake_parent = Box({'node': n.name},default_box=True,box_dots=True)
+      intf_data = create_vlan_member_interface({},vname,fake_parent,topology,0)
+
+      if interface_vlan_mode(intf_data,n,topology) == 'irb':            # We're adding loopback links only for IRB VLANs
+        link_data.interfaces.append(intf_data)
+        topology.links.append(link_data)
+
 """
 get_vlan_data: Get VLAN data structure (node or topology)
 """
@@ -549,6 +616,9 @@ def create_node_vlan(node: Box, vlan: str, topology: Box) -> typing.Optional[Box
       common.fatal(                                                 # ... this should have been detected way earlier
         f'Unknown VLAN {vlan} used on node {node.name}','vlan')
       return None
+    for m in list(node.vlans[vlan].keys()):                         # Remove irrelevant module parameters
+      if not m in node.module and m in topology.module:             # ... it's safe to use direct references, everyone is using VLAN module
+        node.vlans[vlan].pop(m,None)
 
   if not 'mode' in node.vlans[vlan]:                                # Make sure vlan.mode is set
     node.vlans[vlan].mode = get_vlan_mode(node,topology)
@@ -567,8 +637,7 @@ create_svi_interfaces: for every physical interface with access VLAN, create an
 """
 
 def create_svi_interfaces(node: Box, topology: Box) -> dict:
-  global phy_ifattr
-  skip_ifattr = list(phy_ifattr)
+  skip_ifattr = list(topology.defaults.vlan.attributes.phy_ifattr)
   skip_ifattr.extend(topology.defaults.providers.keys())
 
   vlan_ifmap: dict = {}
@@ -576,7 +645,8 @@ def create_svi_interfaces(node: Box, topology: Box) -> dict:
   # VLAN attributes not copied into VLAN interface: we take the global defaults and remove
   # mode attribute from that list as it's needed to set interface VLAN mode
   #
-  svi_skipattr = [ k for k in list(topology.defaults.vlan.vlan_no_propagate or []) if k != "mode" ]
+  svi_skipattr = [ k for k in list(topology.defaults.vlan.attributes.vlan_no_propagate or []) if k != "mode" ] + \
+                  list(topology.defaults.vlan.attributes.vlan_svi_no_propagate)
 
   iflist_len = len(node.interfaces)
   for ifidx in range(0,iflist_len):
@@ -700,9 +770,9 @@ def map_trunk_vlans(node: Box, topology: Box) -> None:
       continue
 
     vlan_list = []
-    for vlan in trunk.keys():
+    for vlan in list(trunk.keys()):
       if not vlan in node.vlans:
-        common.fatal(f'Internal error: VLAN {vlan} should be already defined on node {node.name}')
+        common.fatal(f'Internal error: VLAN {vlan} should be already defined on node {node.name}\n... interface {intf}')
         break
       vlan_list.append(node.vlans[vlan].id)
 
@@ -751,9 +821,9 @@ def remove_vlan_from_trunk(parent_intf: Box, vlan_id: int) -> None:
 rename_vlan_subinterfaces: rename or remove interfaces created from VLAN pseudo-links
 """
 def rename_vlan_subinterfaces(node: Box, topology: Box) -> None:
-  global phy_ifattr
-  skip_ifattr = list(phy_ifattr)
+  skip_ifattr = list(topology.defaults.vlan.attributes.phy_ifattr)
   skip_ifattr.extend(topology.defaults.providers.keys())
+  keep_subif_attr = topology.defaults.vlan.attributes.keep_subif
 
   features = devices.get_device_features(node,topology.defaults)
   if 'switch' in features.vlan.model:                             # No need for VLAN subinterfaces, remove non-routed vlan_member interfaces
@@ -761,6 +831,7 @@ def rename_vlan_subinterfaces(node: Box, topology: Box) -> None:
                           if intf.type != 'vlan_member' or intf.vlan.get('mode','') == 'route' ]
     subif_name = features.vlan.routed_subif_name                  # Just in case: try to get interface name pattern for routed subinterface
 
+  has_vlan_loopbacks = False
   for intf in node.interfaces:                                    # Now loop over remaining vlan_member interfaces
     if intf.type != 'vlan_member':
       continue
@@ -781,6 +852,10 @@ def rename_vlan_subinterfaces(node: Box, topology: Box) -> None:
         f'Internal error: device {node.device} acts as a VLAN-capable {features.vlan.model} but does not have subinterface name template')
       continue
 
+    if not isinstance(intf.parentindex,int):                      # Must be a loopback VLAN interface, skip it
+      has_vlan_loopbacks = True
+      continue
+
     parent_intf = find_parent_interface(intf,node,topology)
     if parent_intf is None:
       common.fatal(f'Internal error: cannot find parent interface for {intf} in node {node.name}')
@@ -816,6 +891,9 @@ def rename_vlan_subinterfaces(node: Box, topology: Box) -> None:
     if intf.vlan.access_id in parent_intf.vlan.get('trunk_id',[]):    # Remove subinterface VLAN from parent interface trunk list
       remove_vlan_from_trunk(parent_intf,intf.vlan.access_id)
 
+  if has_vlan_loopbacks:
+    node.interfaces = [ intf for intf in node.interfaces if intf.parentindex != 'loopback']
+
 """
 If we have a routed native VLAN, and there are no other bridged VLANs left
 in the VLAN trunk, we don't need a trunk at all.
@@ -902,7 +980,8 @@ fix_vlan_gateways -- set VLAN-wide gateway IP
 The link augmentation code sets gateway IP for hosts connected to physical links. That approach does not work
 for VLAN subnets stretched across multiple physical links. We have to fix that here based on host neighbor list.
 
-Please note that we'll have to fix this code when we implement the first-hop gateway module
+Please note that when a first-hop gateway is applied to a VLAN, the link transformation code sets gateway.ipv4
+on every segment of the VLAN, and consequently sets gateway.ipv4 for all hosts attached to a VLAN-enabled link
 """
 def fix_vlan_gateways(topology: Box) -> None:
   for node in topology.nodes.values():
@@ -925,7 +1004,10 @@ def populate_node_vlan_data(n: Box, topology: Box) -> None:
     for vname in n.vlans.keys():                                            # ... to cope with nodes that had VLANs defined
       if vname in topology.get('vlans',{}):                                 # ... through groups.node_data
         topo_data = Box(topology.vlans[vname])                              # Create a copy of topology VLAN
-        topo_data.pop('neighbors',None)                                     # ... and remove neighbors
+        topo_data.pop('neighbors',None)                                     # ... remove neighbors
+        for m in list(topo_data.keys()):                                    # ... and irrelevant module attributes
+          if not m in n.module and m in topology.module:
+            topo_data.pop(m,None)
         n.vlans[vname] = topo_data + n.vlans[vname]                         # ... now merge with the VLAN data
 
 class VLAN(_Module):
@@ -953,14 +1035,20 @@ class VLAN(_Module):
       for vname in node.vlans.keys():
         if node.vlans[vname] is None:
           node.vlans[vname] = {}
-        if 'vlans' in topology and vname in topology.vlans:
-          for kw in ('prefix','id','vni'):
-            if kw in node.vlans[vname]:
-              common.error(
-                f'Cannot set {kw} for VLAN {vname} on node {node.name} -- VLAN is defined globally',
-                common.IncorrectValue,
-                'vlan')
-          node.vlans[vname] = topology.vlans[vname] + node.vlans[vname]
+        if vname in topology.get('vlans',{}):                                     # We have a VLAN defined globally and in a node
+          for kw in ('prefix','id','vni'):                                        # These three attributes MUST NOT be different
+            if not kw in node.vlans[vname]:                                       # OK, attribute not in node VLAN, move on
+              continue
+
+            if kw in topology.vlans[vname] and node.vlans[vname][kw] == topology.vlans[vname][kw]:
+              continue                                                            # Overlap, but the attributes match. OK...
+
+            common.error(
+              f'Cannot set {kw} for VLAN {vname} on node {node.name} -- VLAN is defined globally',
+              common.IncorrectValue,
+              'vlan')
+
+          node.vlans[vname] = topology.vlans[vname] + node.vlans[vname]           # Merge topology VLAN info into node VLAN info
 
     validate_vlan_attributes(node,topology)
 
@@ -987,10 +1075,16 @@ class VLAN(_Module):
         if 'vlan' in intf_node.get('module',[]):                                  # ... is the node a VLAN-aware node?
           intf.vlan = link.vlan + intf.vlan                                       # ... merge link VLAN attributes with interface attributes
 
+    if common.debug_active('vlan'):
+      print(f'VLAN link_pre_transform for {link}')
+      print(f'... VLAN attribute set {v_attr}')
+
     if 'trunk' in v_attr:
+      if not validate_trunk_vlan_list(link):
+        return
       create_vlan_links(link,v_attr,topology)
 
-    svi_skipattr = topology.defaults.vlan.vlan_no_propagate or []                 # VLAN attributes not copied into link data
+    svi_skipattr = topology.defaults.vlan.attributes.vlan_no_propagate or []      # VLAN attributes not copied into link data
     link_vlan = get_link_access_vlan(v_attr)
     routed_vlan = False
     if not link_vlan is None:
@@ -998,7 +1092,8 @@ class VLAN(_Module):
       vlan_data = get_from_box(topology,f'vlans.{link_vlan}')                     # Get global VLAN data
       if isinstance(vlan_data,Box):
         vlan_data = Box({ k:v for (k,v) in vlan_data.items() \
-                                if k not in svi_skipattr })                       # Remove VLAN-specific data
+                                if k not in svi_skipattr },
+                        default_box=True,box_dots=True)                           # Remove VLAN-specific data
         fix_vlan_mode_attribute(vlan_data)                                        # ... and turn mode into vlan.mode
         for (k,v) in vlan_data.items():                                           # Now add the rest to link data
           if not k in link:                                                       # ... have to do the deep merge manually as
@@ -1023,6 +1118,9 @@ class VLAN(_Module):
             intf.ipv4 = False                                                     # ... if so, disable addressing on this interface
             intf.ipv6 = False
 
+  def module_pre_link_transform(self, topology: Box) -> None:
+    create_loopback_vlan_links(topology)
+
   def module_post_link_transform(self, topology: Box) -> None:
     for n in topology.nodes.values():
       if 'vlan' in n.get('module',[]):
